import {SceneEntity} from "./SceneEntity.js";
import {EventDispatcher} from "./EventDispatcher.js";
import {GammaCorrectionShader} from "../postprocessing/GammaCorrection";
import {
    CAMERA_FOV,
    CAMERA_NEAR,
    CAMERA_FAR,
    CAMERA_POSITION,
    ANTIALIAS,
    CLEAR_COLOR,
    MOBILE_CLIENT, SHADOW, DETAIL,
} from "../settings.js";
import {ShadowController} from "../shadow/ShadowController";

function Engine(container, parameters) {
    SceneEntity.call(this, name);

    const self = this;
    const fpsStats = {
        timer: null,
        curr: 0,
        accum: 0,
        framesNum: 0,
        fps: 0
    };
    this.engineInfo = {
        frames: 0,
        prevTime: 0,
        now: 0,
        minDelta: 50,
        value: 0,
        counter: 0,
        values: [],
    };


    this.domelement = document.getElementById(container);

    this.view = new THREE.Scene();

    this.camera = new THREE.PerspectiveCamera(CAMERA_FOV || 45, this.domelement.clientWidth / this.domelement.clientHeight, CAMERA_NEAR, CAMERA_FAR);
    this.camera.position.copy(CAMERA_POSITION);
    this.view.add(this.camera);

    const stateSettings = JSON.parse(localStorage.getItem("AvvyLand.settings"));
    let antialias = ANTIALIAS || false;
    this.postprocess = false;
    this.composer = null;


    const maxRatio = (MOBILE_CLIENT) ? DETAIL.mobile : DETAIL.desktop;
    let pixelRatio = Math.min(window.devicePixelRatio, maxRatio);

    if (stateSettings) {
        antialias = stateSettings.antialiasing;
        pixelRatio = stateSettings.pixelRatio;
    }

    let webGLRendererParameters;
    const defaultParameters = {
        antialias: antialias,
        // preserveDrawingBuffer: true,
    }

    if (!parameters) {
        webGLRendererParameters = defaultParameters;
    } else {
        webGLRendererParameters = Object.assign(defaultParameters, parameters);
    }

    this.renderer = new THREE.WebGLRenderer(webGLRendererParameters);

    this.renderer.outputEncoding = THREE.sRGBEncoding;

    this.renderer.setClearColor(CLEAR_COLOR);
    this.renderer.setSize(this.domelement.clientWidth, this.domelement.clientHeight);

    this.renderer.setPixelRatio(pixelRatio);

    this.renderer.shadowMap.enabled = true;
    this.renderer.shadowMap.type = (MOBILE_CLIENT) ? SHADOW.mobile.type : SHADOW.desktop.type

    this.domelement.append(this.renderer.domElement);


    if (parameters && parameters.postprocess) {
        this.postprocess = true;
        this.composer = new THREE.EffectComposer(this.renderer);
        const renderPass = new THREE.RenderPass(self.view, self.camera);
        this.composer.addPass(renderPass);
        if (parameters.postprocess.name === "outlinePass") {
            this.outlinePass = new THREE.OutlinePass(new THREE.Vector2(this.domelement.clientWidth, this.domelement.clientHeight), self.view, self.camera);
            this.composer.addPass(this.outlinePass);
            this.outlinePass.edgeThickness = parameters.postprocess.edgeThickness ? parameters.postprocess.edgeThickness : 1;
            this.outlinePass.edgeStrength = parameters.postprocess.edgeStrength ? parameters.postprocess.edgeStrength : 1;
            this.outlinePass.edgeGlow = parameters.postprocess.edgeGlow ? parameters.postprocess.edgeGlow : 0;
            this.outlinePass.visibleEdgeColor.set(parameters.postprocess.visibleEdgeColor ? parameters.postprocess.visibleEdgeColor : 0xFFFFFF);
            this.outlinePass.hiddenEdgeColor.set(parameters.postprocess.hiddenEdgeColor ? parameters.postprocess.hiddenEdgeColor : 0xFFFFFF);

            this.effectFXAA = new THREE.ShaderPass(THREE.FXAAShader);
            this.effectFXAA.uniforms['resolution'].value.set(1 / this.domelement.clientWidth, 1 / this.domelement.clientHeight);
            this.composer.addPass(this.effectFXAA);

            const gammaCorrection = new THREE.ShaderPass( GammaCorrectionShader );
            this.composer.addPass(gammaCorrection);

            EventDispatcher.bind("OutlinePass.setSelected", this, function (objects) {
                this.outlinePass.selectedObjects = objects;
            });
        }
    }

    // const clock = new THREE.Clock();

    window.addEventListener("resize", function () {
        self.resize();
    }, false);

    function render() {
        //  const delta = clock.getDelta();
        self.updateFpsCounter();

        if (typeof TWEEN !== "undefined") {
            TWEEN.update();
        }

        let step = 0;
        if (fpsStats.timer) {
            const timeCurr = Date.now();
            step = Math.abs(timeCurr - fpsStats.timer) * 0.001;
            fpsStats.timer = timeCurr;

            fpsStats.accum += step;
            fpsStats.framesNum++;
            if (fpsStats.accum >= 1.0) {
                fpsStats.fps = fpsStats.framesNum;
                fpsStats.accum = 0.0;
                fpsStats.framesNum = 0;
            }
        } else {
            fpsStats.timer = Date.now();
            fpsStats.accum = 0.0;
            fpsStats.framesNum = 0;
        }
        if (step > 0.0 && step < 1.0) {
            self.update(step);
        }

        if (self.postprocess) {
            self.composer.render(step);
        } else {
            self.renderer.render(self.view, self.camera);
        }
        requestAnimationFrame(render);
    }

    EventDispatcher.bind("Engine.togglePostprocess", this, function (bool) {
        self.postprocess = bool;
    });
    EventDispatcher.bind("Engine.get.scene", this, function () {
        return self.view;
    });
    EventDispatcher.bind("Engine.get.camera", this, function () {
        return self.camera;
    });
    EventDispatcher.bind("Engine.get.renderer", this, function () {
        return self.renderer;
    });
    EventDispatcher.bind("Engine.get.engineInfo", this, function () {
        this.engineInfo.geometries = this.renderer.info.memory.geometries;
        this.engineInfo.textures = this.renderer.info.memory.textures;
        this.engineInfo.calls = this.renderer.info.render.calls;
        this.engineInfo.triangles = this.renderer.info.render.triangles;
        this.engineInfo.programs = this.renderer.info.programs.length;
        return this.engineInfo;
    });
    EventDispatcher.bind("Engine.add", this, this.add);
    EventDispatcher.bind("Engine.remove", this, this.remove);
    EventDispatcher.bind("Engine.resize", this, this.resize);

    const props = (MOBILE_CLIENT) ? SHADOW.mobile : SHADOW.desktop;
    this.shadowController = new ShadowController(props);

    render();
}

Engine.prototype = Object.create(SceneEntity.prototype);
Engine.constructor = Engine;

Engine.prototype.resize = function () {
    this.camera.aspect = this.domelement.clientWidth / this.domelement.clientHeight;
    this.camera.updateProjectionMatrix();
    this.renderer.setSize(this.domelement.clientWidth, this.domelement.clientHeight);
    if (this.composer) {
        this.composer.setSize(this.domelement.clientWidth, this.domelement.clientHeight);
    }
    if (this.effectFXAA) {
        this.effectFXAA.uniforms['resolution'].value.set(1 / this.domelement.clientWidth, 1 / this.domelement.clientHeight);
    }
};

Engine.prototype.updateFpsCounter = function () {
    this.engineInfo.frames++;
    this.engineInfo.now = Date.now();

    if (this.engineInfo.now >= this.engineInfo.prevTime + 1000) {
        const value = Math.round((this.engineInfo.frames * 1000) / (this.engineInfo.now - this.engineInfo.prevTime));
        this.engineInfo.value = value;
        this.engineInfo.prevTime = this.engineInfo.now;
        this.engineInfo.frames = 0;
    }

};
export {Engine};
